from __future__ import print_function

import os
from copy import copy
from enum import Enum

from .. import CatBoostError, CatBoost
from .evaluation_result import EvaluationResults, MetricEvaluationResult
from ._fold_models_handler import FoldModelsHandler
from ._readers import _SimpleStreamingFileReader
from ._splitter import _Splitter
from .execution_case import ExecutionCase
from .factor_utils import LabelMode, FactorUtils


class EvalType(Enum):
    """
        Type of feature evaluation:
            All: All factors presented
            SeqRem:  Each factor while other presented
            SeqAdd:  Each factor while other removed
            SeqAddAndAll:  SeqAdd + All
    """
    All = 'All'
    SeqRem = 'SeqRem'
    SeqAdd = 'SeqAdd'
    SeqAddAndAll = 'SeqAddAndAll'


class CatboostEvaluation(object):

    def __init__(self,
                 path_to_dataset,
                 fold_size,
                 fold_count,
                 column_description,
                 fold_offset=0,
                 group_column=None,
                 working_dir=None,
                 remove_models=True,
                 delimiter='\t',
                 has_header=False,
                 partition_random_seed=0,
                 min_fold_count=1):
        """
        Args:
            :param path_to_dataset: (str) Path to the dataset to be used for evaluation.
            :param fold_size: (int) Size of the folds in cross-validation.
            :param fold_count: (int) Number of times we get a new fold, learn a model and check results as if
            there wouldn't be any offset. If there'are some offset it means that the real count of folds will
            be smaller.
            :param column_description: (str) Path to the file where column description is placed.
            :param fold_offset: (int) Number of folds we skip before begin to make cross-validation.
            :param group_column: (int) GroupId column index in the dataset file.
            'None' value means absence of grouping information in the dataset (it's the default).
            :param working_dir: Working dir for temporary files
            :param remove_models: (bool) Set true if you want models to be removed after applying them.
            :param delimiter: (str) Field delimiter used in dataset files.
            :param has_header: (bool) Set true if you want to skip first line in dataset files.
            :param partition_random_seed: (int) The seed for random value generator used for getting permutations for
             cross-validation.
            :param min_fold_count: (int) Minimun amount of folds dataset can be cut to.
        """
        import os.path

        self._current_dir = os.getcwd()
        self._path_to_dataset = os.path.join(self._current_dir, path_to_dataset)
        self._column_description = os.path.join(self._current_dir,
                                                column_description) if column_description is not None else None
        self._fold_offset = fold_offset
        self._fold_count = fold_count
        self._fold_size = fold_size
        self._delimiter = delimiter
        self._has_header = has_header
        self._seed = partition_random_seed
        self._min_fold_count = int(min_fold_count)
        self._remove_models = remove_models

        if group_column is not None:
            self._group_feature_num = int(group_column)
        else:
            self._group_feature_num = group_column

        if working_dir is None:
            import tempfile
            self._working_dir = tempfile.mkdtemp()
        else:
            self._working_dir = working_dir

    def __go_to_working_dir(self):
        current = os.getcwd()
        os.chdir(self._working_dir)
        return current

    @staticmethod
    def _create_eval_feature_cases(params, features_to_eval, eval_type, label_mode):
        if len(features_to_eval) == 0:
            raise CatBoostError("Provide at least one feature to evaluation")

        # baseline
        test_cases = list()
        baseline_case = ExecutionCase(params,
                                      ignored_features=list(features_to_eval),
                                      label=FactorUtils.create_label(features_to_eval,
                                                                     features_to_eval,
                                                                     label_mode=label_mode)
                                      )
        # test
        if eval_type == EvalType.All or eval_type == EvalType.SeqAddAndAll or len(features_to_eval) == 1:
            test_cases.append(ExecutionCase(params,
                                            ignored_features=[],
                                            label=FactorUtils.create_label(features_to_eval,
                                                                           [],
                                                                           label_mode=label_mode)
                                            ))
        elif eval_type == EvalType.SeqRem:
            for feature_num in features_to_eval:
                test_cases.append(ExecutionCase(params,
                                                ignored_features=[feature_num],
                                                label=FactorUtils.create_label(features_to_eval,
                                                                               [feature_num],
                                                                               label_mode=label_mode)
                                                ))
        elif eval_type == EvalType.SeqAdd or eval_type == EvalType.SeqAddAndAll:
            for feature_num in features_to_eval:
                cur_features = copy(features_to_eval)
                cur_features.remove(feature_num)
                test_cases.append(ExecutionCase(params,
                                                label=FactorUtils.create_label(features_to_eval,
                                                                               cur_features,
                                                                               label_mode=label_mode),
                                                ignored_features=list(cur_features)))
        elif eval_type != EvalType.All:
            raise AttributeError("Don't support {} mode.", eval_type.value)
        return baseline_case, test_cases

    @staticmethod
    def _create_evaluation_results(by_case_results):
        group_by_metric = dict()

        for (case, case_result) in by_case_results.items():
            for (metric, evaluation_result) in case_result.items():
                if metric not in group_by_metric:
                    group_by_metric[metric] = list()
                group_by_metric[metric].append(evaluation_result)

        results = list()
        for (metric, metric_results) in group_by_metric.items():
            results.append(MetricEvaluationResult(metric_results))
        return EvaluationResults(results)

    def get_working_dir(self):
        return self._working_dir

    def _calculate_result_metrics(self, cases, metrics, thread_count=-1, evaluation_step=1):
        """
        This method calculate metrics and return them.

        Args:
            :param cases: List of the ExecutionCases you want to evaluate
            :param metrics: List of the metrics to be computed
            :param thread_count: Count of threads to use.
            :param: evaluation_step: Step to evaluate metrics
            :return: instance of EvaluationResult
        """
        cases_set = set(cases)
        if len(cases_set) != len(cases):
            raise CatBoostError("Found duplicate cases in " + cases)
        current_wd = self.__go_to_working_dir()
        try:
            if self._fold_count <= self._fold_offset:
                error_msg = 'Count of folds(folds_count - offset) need to be at least one: offset {}, folds_count {}.'
                raise AttributeError(error_msg.format(self._fold_offset,
                                                      self._fold_count))

            handler = FoldModelsHandler(cases=cases,
                                        metrics=metrics,
                                        eval_step=evaluation_step,
                                        thread_count=thread_count,
                                        remove_models=self._remove_models)

            reader = _SimpleStreamingFileReader(self._path_to_dataset,
                                                sep=self._delimiter,
                                                has_header=self._has_header,
                                                group_feature_num=self._group_feature_num)
            splitter = _Splitter(reader,
                                 self._column_description,
                                 seed=self._seed,
                                 min_folds_count=self._min_fold_count)

            result = handler.proceed(splitter=splitter,
                                     fold_size=self._fold_size,
                                     folds_count=self._fold_count,
                                     fold_offset=self._fold_offset)

            return self._create_evaluation_results(result)
        finally:
            os.chdir(current_wd)

    def eval_features(self,
                      learn_config,
                      features_to_eval,
                      loss_function=None,
                      eval_type=EvalType.SeqAdd,
                      eval_metrics=None,
                      thread_count=-1,
                      eval_step=None,
                      label_mode=LabelMode.AddFeature):
        """ Evaluate features.
            Args:
            learn_config: dict with params or instance of CatBoost. In second case instance params will be used
            features_to_eval: list of indices of features to evaluate
            loss_function: one of CatBoost loss functions, get it from learn_config if not specified
            eval_type: Type of feature evaluate (All, SeqAdd, SeqRem)
            eval_metrics: Additional metrics to calculate
            thread_count: thread_count to use. If not none will override learn_config values
            Returns
            -------
            result : Instance of EvaluationResult class
        """
        features_to_eval = set(features_to_eval)
        if eval_metrics is None:
            eval_metrics = []
        eval_metrics = eval_metrics if isinstance(eval_metrics, list) else [eval_metrics]
        if isinstance(learn_config, CatBoost):
            params = learn_config.get_params()
        else:
            params = dict(learn_config)

        if loss_function is not None:
            if "loss_function" in params and params["loss_function"] != loss_function:
                raise CatBoostError("Loss function in params {} should be equal to feature evaluation objective "
                                    "function {}".format(params["loss_function"], loss_function))
        else:
            if "loss_function" not in params:
                raise CatBoostError("Provide loss function in params or as option to eval_features method")

        if thread_count is not None and thread_count != -1:
            params["thread_count"] = thread_count

        if eval_step is None:
            eval_step = 1

        if loss_function is not None:
            params["loss_function"] = loss_function
        else:
            loss_function = params["loss_function"]

        if params["loss_function"] == "PairLogit":
            raise CatBoostError("Pair classification is not supported")

        baseline_case, test_cases = self._create_eval_feature_cases(params,
                                                                    features_to_eval,
                                                                    eval_type=eval_type,
                                                                    label_mode=label_mode)
        if loss_function not in eval_metrics:
            eval_metrics.append(loss_function)

        return self.eval_cases(baseline_case=baseline_case,
                               compare_cases=test_cases,
                               eval_metrics=eval_metrics,
                               thread_count=thread_count,
                               eval_step=eval_step)

    def eval_cases(self,
                   baseline_case,
                   compare_cases,
                   eval_metrics,
                   thread_count=-1,
                   eval_step=1):
        """More flexible evaluation of any cases.
            Args:
            baseline_case: Execution case used for baseline
            compare_cases: List of cases to compare
            eval_metrics: Metrics to calculate
            thread_count: thread_count to use.  Will override one in cases
            Returns
            -------
            result : Instance of EvaluationResult class
        """
        if not isinstance(compare_cases, list):
            compare_cases = [compare_cases]

        cases = [baseline_case]
        cases += compare_cases

        for case in cases:
            case._set_thread_count(thread_count)

        metric_result = self._calculate_result_metrics(cases,
                                                       eval_metrics,
                                                       thread_count=thread_count,
                                                       evaluation_step=eval_step)
        metric_result.set_baseline_case(baseline_case)
        return metric_result



from .hnsw import EDistance, EVectorComponentType, Pool, Hnsw, HnswException, transform_mobius, HnswEstimator, OnlineHnsw
__all__ = ['EDistance', 'EVectorComponentType', 'Pool', 'Hnsw', 'HnswException', 'transform_mobius', 'HnswEstimator', 'OnlineHnsw']



from __future__ import print_function

import os
import time

from .. import CatBoostError
from ..eval.log_config import get_eval_logger
from ..eval.utils import make_dirs_if_not_exists
from ..eval.evaluation_result import CaseEvaluationResult
from ._fold_model import FoldModel


class FoldModelsHandler(object):
    """
    Class that is responsible for learning models and computing their metrics
    """
    """
       All models are placed to the default directory "models".
       """
    __MODEL_DIR = 'models'

    @staticmethod
    def _remove_model_dir():
        try:
            if os.path.exists(FoldModelsHandler.__MODEL_DIR):
                os.rmdir(FoldModelsHandler.__MODEL_DIR)
        except OSError as err:
            get_eval_logger().warning(str(err))

    def __init__(self, metrics, cases, thread_count, eval_step, remove_models):
        """
        Args:
            :param remove_models: Set true if you want models to be removed after applying them.

        """
        self._cases = cases
        self._metrics = metrics

        self._case_results = dict()
        for case in self._cases:
            self._case_results[case] = dict()

        self._thread_count = thread_count
        self._eval_step = eval_step
        self._flag_remove_models = remove_models
        self._metric_descriptions = None

    def _init_case_results(self, metric_descriptions):
        self._metric_descriptions = metric_descriptions
        for case in self._cases:
            case_result = self._case_results[case]
            for metric_description in self._metric_descriptions:
                case_result[metric_description] = CaseEvaluationResult(case,
                                                                       metric_description,
                                                                       eval_step=self._eval_step)

    def _compute_metrics(self, metrics, grouped_by_case_models, learn_folds, skipped_folds, rest_folds):
        metric_calcers = {}

        for case, case_models in grouped_by_case_models.items():
            metric_calcers[case] = list()
            for case_model in case_models:

                metric_calcer = case_model.create_metrics_calcer(metrics,
                                                                 eval_step=self._eval_step,
                                                                 thread_count=self._thread_count)
                metric_calcers[case].append(metric_calcer)

                if self._metric_descriptions is None:
                    self._init_case_results(metric_calcer.metric_descriptions())
                elif self._metric_descriptions != metric_calcer.metric_descriptions():
                    raise CatBoostError("Error: metric names should be consistent")

        for file_num, fold_file in enumerate(learn_folds + skipped_folds + rest_folds):
            pool = FoldModelsHandler._create_pool(fold_file, self._thread_count)

            for case, case_models in grouped_by_case_models.items():
                calcers = metric_calcers[case]

                for model_num, model in enumerate(case_models):
                    if file_num != model_num:
                        calcers[model_num].add(pool)

        for case, case_models in grouped_by_case_models.items():
            calcers = metric_calcers[case]
            case_results = self._case_results[case]
            for calcer, model in zip(calcers, case_models):
                scores = calcer.eval_metrics()
                for metric in self._metric_descriptions:
                    case_results[metric]._add(model, scores.get_result(metric))

    @staticmethod
    def _fit_model(pool, case, fold_id, model_path):
        from .. import CatBoost
        # Learn model
        make_dirs_if_not_exists(FoldModelsHandler.__MODEL_DIR)

        feature_count = pool.num_col()
        if "ignored_features" in case.get_params():
            ignored_features = case.get_params()["ignored_features"]
            if len(ignored_features) and max(ignored_features) >= feature_count:
                raise CatBoostError("Error: input parameter contains feature indices wich are not available in pool: "
                                    "{}\n "
                                    "Check eval_feature set and ignored features options".format(ignored_features))
        get_eval_logger().debug('Learn model {} on fold #{}'.format(str(case), fold_id))
        cur_time = time.time()
        instance = CatBoost(params=case.get_params())
        instance.fit(pool)
        instance.save_model(fname=model_path)

        get_eval_logger().debug('Operation was done in {} seconds'.format(time.time() - cur_time))
        return FoldModel(case, model_path, fold_id)

    def _fit_models(self, learn_files, fold_id_bias):
        """
        Train models for each algorithm and learn dataset(folds). Than return them.

        Args:
            :param learn_files: Entities of FoldStorage for learning models.
            :return: Dictionary of models where the key is case and the value is models on learn folds
        """
        make_dirs_if_not_exists(FoldModelsHandler.__MODEL_DIR)

        models = {}
        for case in self._cases:
            models[case] = list()

        for file_num, learn_file in enumerate(learn_files):
            pool = FoldModelsHandler._create_pool(learn_file, self._thread_count)
            fold_id = fold_id_bias + file_num

            for case in self._cases:
                model_path = os.path.join(FoldModelsHandler.__MODEL_DIR,
                                          FoldModelsHandler._create_model_name(case, fold_id))
                get_eval_logger().debug("For model {} on fold #{} path is {}".format(str(case), fold_id, model_path))
                fold_model = self._fit_model(pool, case, fold_id, model_path)
                get_eval_logger().info("Model {} on fold #{} was fitted".format(str(case), fold_id))
                models[case].append(fold_model)

        return models

    def proceed(self, splitter, fold_size, folds_count, fold_offset):
        """
        Run all processes to gain stats. It applies algorithms to fold files that gains from learning. It keeps
        stats inside models and models are stored in DataFrame. Columns are matched to the different algos and rows to
        the folds.

        Args:
            :param splitter: Splitter entity.
            :param fold_size: The size of fold.
            :param folds_count: Count of golds.
            :param fold_offset: The offset (count of folds that we want to skip).
            :return: return dict: keys metric to CaseEvaluationResult

        """
        try:
            folds_sets = splitter.create_fold_sets(fold_size, folds_count)
            fold_groups_files = splitter.fold_groups_files_generator(folds_sets,
                                                                     fold_offset)
            fold_id_bias = fold_offset

            for learn_folds, skipped_folds, rest_folds in fold_groups_files:
                if len(learn_folds) == 0:
                    continue
                list_models = []
                try:
                    permutation_models = self._fit_models(learn_folds, fold_id_bias)
                    for case, case_models in permutation_models.items():
                        list_models += case_models

                    learn_folds_count = len(learn_folds)
                    get_eval_logger().info("Start metric computation for folds [{}, {})"
                                           .format(fold_id_bias, fold_id_bias + learn_folds_count))
                    self._compute_metrics(self._metrics,
                                          permutation_models,
                                          learn_folds, skipped_folds, rest_folds)
                    get_eval_logger().info("Computation of metrics for  folds [{}, {}) is completed"
                                           .format(fold_id_bias, fold_id_bias + learn_folds_count))
                    fold_id_bias += learn_folds_count
                finally:
                    # Do it each step because don't want to occupy a lot of memory
                    splitter.clean_folds()
                    self._remove_models(list_models)

        finally:
            # Also sometimes we need to remove models and always need to try to remove folds directory.
            self._clean(splitter)

        return self._case_results

    def _clean(self, splitter):
        if self._flag_remove_models:
            self._remove_model_dir()
        splitter.clean()

    @staticmethod
    def _remove_models(list_models):
        get_eval_logger().debug('Remove models {}'.format(str(list_models)))
        for model in list_models:
            model.delete()

    @staticmethod
    def _create_model_name(model_case, fold):
        import uuid
        id_str = str(uuid.uuid1()).replace("-", "_")
        model_name = "model_{_name}_fold_{_fold}_{_uuid}.bin".format(_name=model_case.get_label(),
                                                                     _fold=fold,
                                                                     _uuid=id_str)
        return model_name

    @staticmethod
    def _create_pool(fold_file, thread_count=-1):
        from .. import Pool
        data_pool = Pool(fold_file.path(),
                         column_description=fold_file.column_description(),
                         delimiter=fold_file.get_separator(),
                         thread_count=thread_count)
        return data_pool



from enum import Enum


class LabelMode(Enum):
    AddFeature = "Feature"
    IgnoreFeature = "Ignore"


class FactorUtils(object):

    @staticmethod
    def extract_factors(string_factors):
        string_factors = string_factors.strip()
        if not string_factors:
            return {}

        list_factors = []
        eval_factor_units = string_factors.split(';')
        for eval_factor_unit in eval_factor_units:
            try:
                factor = int(eval_factor_unit)
                list_factors.append(factor)
            except ValueError:
                string_bounds = eval_factor_unit.split('-')
                if len(string_bounds) != 2:
                    raise AttributeError('Range need to contain exactly two numbers!')
                begin_range = int(string_bounds[0])
                end_range = int(string_bounds[1])
                list_factors += list(range(begin_range, end_range + 1))

        return set(list_factors)

    @staticmethod
    def factors_to_string(factors):
        if len(factors) == 0:
            return ''
        parts = []
        factors_list = sorted(factors)
        begin = factors_list[0]
        for i in range(1, len(factors_list)):
            if factors_list[i] != factors_list[i - 1] + 1:
                end = factors_list[i - 1]
                if begin != end:
                    parts.append('{}-{}'.format(begin, end))
                else:
                    parts.append(str(begin))
                begin = factors_list[i]
        end = len(factors_list) - 1
        if begin != factors_list[end]:
            parts.append('{}-{}'.format(begin, factors_list[end]))
        else:
            parts.append(str(begin))
        return ';'.join(parts)

    @staticmethod
    def compress_string_factors(string_factors):
        factors = FactorUtils.extract_factors(string_factors)
        compressed_string_factors = FactorUtils.factors_to_string(factors)
        return compressed_string_factors

    @staticmethod
    def single_range_to_string(left, right):
        if left != right:
            return "{}-{}".format(left, right)
        else:
            return "{}".format(left)

    @staticmethod
    def group_factors_by_range(factors_set):
        factors = sorted(list(factors_set))
        is_start = []
        for i in range(0, len(factors)):
            is_start.append(i == 0 or (factors[i] != factors[i - 1] + 1))

        grouped_factors = []
        i = 0
        while i < len(factors):
            if is_start[i]:
                grouped_factors.append([])
            grouped_factors[-1].append(factors[i])
            i += 1
        return grouped_factors

    @staticmethod
    def factors_to_ranges_string(factors_set):
        if factors_set is None or len(factors_set) == 0:
            return "None"
        grouped_factors = FactorUtils.group_factors_by_range(factors_set)

        return ';'.join([FactorUtils.single_range_to_string(min(x), max(x)) for x in grouped_factors])

    @staticmethod
    def create_label(all_eval_features, removed_features, label_mode):
        eval_features = set(all_eval_features)
        if label_mode == LabelMode.AddFeature:
            add_features = eval_features - set(removed_features)
            return "Features: {}".format(FactorUtils.factors_to_ranges_string(add_features))
        else:
            return "Ignore: {}".format(FactorUtils.factors_to_ranges_string(set(removed_features)))



"""
This class provides the abstraction of model. I.e. it is an object relevant to particular learn fold and particular
algorithm with parameters.
"""

import os
from .. import CatBoostError, CatBoost


class FoldModel:

    def __init__(self, case, model_path, model_id):
        self._id = model_id
        self._case = case
        self._model_path = model_path

    def __repr__(self):
        return 'Model_{}_id_{}'.format(str(self._case), self._id)

    def create_metrics_calcer(self, metrics, thread_count, eval_step=1):
        if not os.path.exists(self._model_path):
            raise CatBoostError("Model was deleted. Can't create calcer now")
        model = CatBoost()
        model.load_model(self._model_path)
        return model.create_metric_calcer(metrics, thread_count=thread_count, eval_period=eval_step)

    def get_case(self):
        return self._case

    def get_fold_id(self):
        return self._id

    def delete(self):
        if os.path.exists(self._model_path):
            os.remove(self._model_path)



#  tests for catboost-1.2-py39h06a4308_0 (this is a generated file);
print('===== testing package: catboost-1.2-py39h06a4308_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import subprocess
import pkgutil
import platform

import catboost
import numpy as np

from catboost import Pool, CatBoostRegressor
from catboost.datasets import adult
from catboost.text_processing import Tokenizer


py_impl = platform.python_implementation().lower()
machine = platform.machine().lower()

print("Python implementation:", py_impl)
print("              Machine:", machine)

subprocess.run(["pip", "check"])


# Tokenizer
text="Still, I would love to see you at 12, if you don't mind"

tokenized = Tokenizer(lowercasing=True,
                    separator_type='BySense',
                    token_types=['Word', 'Number']).tokenize(text)
print("Tokenized text:\n")
print(tokenized)

# CatboostRegressor

# initialize data
train_data = np.random.randint(0, 
                               100, 
                               size=(100, 10))
train_label = np.random.randint(0, 
                                1000, 
                                size=(100))
test_data = np.random.randint(0, 
                              100, 
                              size=(50, 10))
# initialize Pool
train_pool = Pool(train_data, 
                  train_label, 
                  cat_features=[0,2,5])
test_pool = Pool(test_data, 
                 cat_features=[0,2,5]) 

# specify the training parameters 
model = CatBoostRegressor(iterations=2, 
                          depth=2, 
                          learning_rate=1, 
                          loss_function='RMSE')
#train the model
model.fit(train_pool)
# make the prediction using the resulting model
preds = model.predict(test_pool)

print("CatboostRegressor:\n")
print(preds)
#  --- run_test.py (end) ---

print('===== catboost-1.2-py39h06a4308_0 OK =====');



import sys
import imp
import os
from six import iteritems
from enum import IntEnum
from contextlib import contextmanager
import json


def get_so_paths(dir_name):
    dir_name = os.path.join(os.path.dirname(__file__), dir_name)
    list_dir = os.listdir(dir_name) if os.path.isdir(dir_name) else []
    return [os.path.join(dir_name, so_name) for so_name in list_dir if so_name.split('.')[-1] in ['so', 'pyd']]


def get_hnsw_bin_module():
    if '_hnsw' in sys.modules:
        return sys.modules['_hnsw']
    so_paths = get_so_paths('./')
    for so_path in so_paths:
        try:
            loaded_hnsw = imp.load_dynamic('_hnsw', so_path)
            sys.modules['hnsw._hnsw'] = loaded_hnsw
            return loaded_hnsw
        except ImportError:
            pass
    from . import _hnsw
    return _hnsw


@contextmanager
def log_fixup():
    _hnsw._set_logger(sys.stdout)
    try:
        yield
    finally:
        _hnsw._reset_logger()


class EDistance(IntEnum):
    DotProduct = 0
    L1 = 1
    L2Sqr = 2


class EVectorComponentType(IntEnum):
    Float = 0
    I8 = 1
    I32 = 2

_hnsw = get_hnsw_bin_module()

HnswException = _hnsw.HnswException

_DenseVectorStorage = {
    EVectorComponentType.Float: _hnsw._DenseFloatVectorStorage,
    EVectorComponentType.I8: _hnsw._DenseI8VectorStorage,
    EVectorComponentType.I32: _hnsw._DenseI32VectorStorage
}

_HnswDenseVectorIndex = {
    EVectorComponentType.Float: _hnsw._HnswDenseFloatVectorIndex,
    EVectorComponentType.I8: _hnsw._HnswDenseI8VectorIndex,
    EVectorComponentType.I32: _hnsw._HnswDenseI32VectorIndex
}

_transform_mobius = {
    EVectorComponentType.Float: _hnsw._transform_mobius_float,
    EVectorComponentType.I8: _hnsw._transform_mobius_i8,
    EVectorComponentType.I32: _hnsw._transform_mobius_i32
}

_OnlineHnswDenseVectorIndex = {
    EVectorComponentType.Float: _hnsw._OnlineHnswDenseFloatVectorIndex,
    EVectorComponentType.I8: _hnsw._OnlineHnswDenseI8VectorIndex,
    EVectorComponentType.I32: _hnsw._OnlineHnswDenseI32VectorIndex,
}


class Pool:
    """
    Pool is a storage of vectors
    """

    def __init__(self, vectors_path, dtype, dimension, vectors_bin_data=None):
        """
        Pool is a storage of vectors. You can create it from row-major binary file or
        binary data of vectors.

        Parameters
        ----------
        vectors_path : string
            Path to binary file with vectors.

        dtype : EVectorComponentType
            Type of vectors.

        dimension : int
            Dimension of vectors.

        vectors_bin_data : bytes
            Binary data of vectors.
        """
        self.vectors_path = vectors_path
        self.dtype = dtype
        self.dimension = dimension
        assert (vectors_bin_data is None) ^ (vectors_path is None)
        if vectors_path is not None:
            self._storage = _DenseVectorStorage[dtype](vectors_path, dimension)
            self._data = None
        if vectors_bin_data is not None:
            self._storage = _DenseVectorStorage[dtype](None, dimension, vectors_bin_data)
            self._data = vectors_bin_data

    @classmethod
    def from_file(cls, vectors_path, dtype, dimension):
        """
        Create pool from binary file.

        Parameters
        ----------
        vectors_path : string
            Path to binary file with vectors.

        dtype : EVectorComponentType
            Type of vectors.

        dimension : int
            Dimension of vectors.
        """
        return Pool(vectors_path, dtype, dimension, None)

    @classmethod
    def from_bytes(cls, vectors_bin_data, dtype, dimension):
        """
        Create pool from binary data.

        Parameters
        ----------
        vectors_bin_data : bytes
            Binary data of vectors.

        dtype : EVectorComponentType
            Type of vectors.

        dimension : int
            Dimension of vectors.
        """
        return Pool(None, dtype, dimension, vectors_bin_data)

    def get_item(self, id):
        """
        Get item from storage by id.

        Parameters
        ----------
        id : int
            Index of item in storage.

        Returns
        -------
        item : numpy.ndarray
        """
        return self._storage._get_item(id)

    def get_num_items(self):
        """
        Get the number of items in storage.

        Returns
        -------
        num_items : int
        """
        return self._storage._get_num_items()


def transform_mobius(pool):
    """
    Transform pool for fast dot product search on HNSW graph
    https://papers.nips.cc/paper/9032-mobius-transformation-for-fast-inner-product-search-on-graph.pdf

    Parameters
    ----------
    pool : Pool

    Returns
    -------
    transformed_pool : Pool
    """
    transformed_pool = Pool.from_bytes(bytes(0), EVectorComponentType.Float, pool.dimension)
    transformed_pool._storage = _transform_mobius[pool.dtype](pool._storage)
    return transformed_pool


class Hnsw:
    """
    Class for building, loading and working with Hierarchical Navigable Small World index.
    """
    def __init__(self):
        """
        Create object for working with HNSW.
        """
        self._index = None
        self._data = None

    def build(self, pool, distance, max_neighbors=None, search_neighborhood_size=None, num_exact_candidates=None,
              batch_size=None, upper_level_batch_size=None, level_size_decay=None, num_threads=None, verbose=False,
              report_progress=True, snapshot_file=None, snapshot_interval=None):
        """
        Build index with given options.

        Parameters
        ----------
        pool : Pool
            Pool of vectors for which index will be built.

        distance : EDistance
            Distance that should be used for finding nearest vectors.

        max_neighbors : int (default=32)
            Maximum number of neighbors that every item can be connected with.

        search_neighborhood_size : int (default=300)
            Search neighborhood size for ANN-search.
            Higher values improve search quality in expense of building time.

        num_exact_candidates : int (default=100)
            Number of nearest vectors to take from batch.
            Higher values improve search quality in expense of building time.

        batch_size : int (default=1000)
            Number of items that added to graph on each step of algorithm.

        upper_level_batch_size : int (default=40000)
            Batch size for building upper levels.

        level_size_decay : int (default=max_neighbors/2)
            Base of exponent for decaying level sizes.

        num_threads : int (default=number of CPUs)
            Number of threads for building index.

        report_progress : bool (default=True)
            Print progress of building.

        verbose : bool (default=False)
            Print additional information about time of building.

        snapshot_file : string (default=None)
            Path for saving snapshots during the index building.

        snapshot_interval : int (default=600)
            Interval between saving snapshots (seconds).
            Snapshot is saved after building each level also.
        """
        params = {}
        not_params = ["not_params", "self", "params", "__class__", "pool", "distance"]
        for key, value in iteritems(locals()):
            if key not in not_params and value is not None:
                params[key] = value
        self._index = _HnswDenseVectorIndex[pool.dtype](pool._storage, distance)
        with log_fixup():
            self._index._build(json.dumps(params))

    def _check_index(self):
        if self._index is None:
            raise HnswException("Index is not built and not loaded")

    def save(self, index_path):
        """
        Save index to file.

        Parameters
        ----------
        index_path : string
            Path to file for saving index.
        """
        self._check_index()
        self._index._save(index_path)

    def load(self, index_path, pool, distance):
        """
        Load index from file.

        Parameters
        ----------
        index_path : string
            Path to file for loading index.

        pool : Pool
            Pool of vectors for which index will be loaded.

        distance : EDistance
            Distance that should be used for finding nearest vectors.
        """
        self._index = _HnswDenseVectorIndex[pool.dtype](pool._storage, distance)
        self._index._load(index_path)
        self._data = None

    def load_from_bytes(self, index_data, pool, distance):
        """
        Load index from bytes.

        Parameters
        ----------
        index_data : bytes
            Index binary data.

        pool : Pool
            Pool of vectors for which index will be loaded.

        distance : EDistance
            Distance that should be used for finding nearest vectors.
        """
        self._index = _HnswDenseVectorIndex[pool.dtype](pool._storage, distance)
        self._index._load_from_bytes(index_data)
        self._data = index_data

    def get_nearest(self, query, top_size, search_neighborhood_size, distance_calc_limit=0):
        """
        Get approximate nearest neighbors for query from index.

        Parameters
        ----------
        query : list or numpy.ndarray
            Vector for which nearest neighbors should be found.

        top_size : int
            Required number of neighbors.

        search_neighborhood_size : int
            Search neighborhood size for ANN-search.
            Higher values improve search quality in expense of search time.
            It should be equal or greater than top_size.

        distance_calc_limit : int (default=0)
            Limit of distance calculation.
            To guarantee satisfactory search time at the expense of quality.
            0 is equivalent to no limit.

        Returns
        -------
        neighbors : list of tuples (id, distance)
        """
        self._check_index()
        return self._index._get_nearest(query, top_size, search_neighborhood_size, distance_calc_limit)


class HnswEstimator:
    """
    Class for building, loading and working with Hierarchical Navigable Small World index with SciKit-Learn
    Estimator compatible interface.
    Mostly drop-in replacement for sklearn.neighbors.NearestNeighbors (except for some parameters)
    """

    def __init__(self, n_neighbors=5,
                 distance=EDistance.DotProduct, max_neighbors=32, search_neighborhood_size=300,
                 num_exact_candidates=100, batch_size=1000, upper_level_batch_size=40000,
                 level_size_decay=None):
        """
        Parameters
        ----------
        n_neighbors : int, default=5
            Number of neighbors to use by default for kneighbors queries.


        distance : EDistance
            Distance that should be used for finding nearest vectors.

        max_neighbors : int (default=32)
            Maximum number of neighbors that every item can be connected with.

        search_neighborhood_size : int (default=300)
            Search neighborhood size for ANN-search.
            Higher values improve search quality in expense of building time.

        num_exact_candidates : int (default=100)
            Number of nearest vectors to take from batch.
            Higher values improve search quality in expense of building time.

        batch_size : int (default=1000)
            Number of items that added to graph on each step of algorithm.

        upper_level_batch_size : int (default=40000)
            Batch size for building upper levels.

        level_size_decay : int (default=max_neighbors/2)
            Base of exponent for decaying level sizes.
        """
        for key, value in iteritems(locals()):
            if key not in ['self', '__class__']:
                setattr(self, key, value)

    def _check_index(self):
        if self._index is None:
            raise HnswException("Index is not built and not loaded")

    def fit(self, X, y=None, num_threads=None, verbose=False, report_progress=True, snapshot_file=None,
            snapshot_interval=600):
        """
        Fit the HNSW model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_values)

        y: None
            Added to be compatible with Estimator API

        num_threads : int (default=number of CPUs)
            Number of threads for building index.

        report_progress : bool (default=True)
            Print progress of building.

        verbose : bool (default=False)
            Print additional information about time of building.

        snapshot_file : string (default=None)
            Path for saving snapshots during the index building.

        snapshot_interval : int (default=600)
            Interval between saving snapshots (seconds).

        Returns
        -------
        model : HnswEstimator

        """
        self._index, self._index_data = _hnsw._init_index(X, self.distance)

        params = self._get_params(return_none=False)
        not_params = ["not_params", "self", "params", "__class__", "X", "y"]
        for key, value in iteritems(locals()):
            if key not in not_params and value is not None:
                params[key] = value
        del params['distance']

        with log_fixup():
            self._index._build(json.dumps(params))
        return self

    def _get_params(self, return_none):
        params = {}
        for key, value in self.__dict__.items():
            if key[0] != '_' and (return_none or (value is not None)):
                params[key] = value
        return params

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.
        """
        return self._get_params(return_none=True)

    def set_params(self, **params):
        """
        Set the parameters of this estimator.

        Parameters
        ----------
        **params : dict
            HnswEstimator parameters.

        Returns
        -------
        self : HnswEstimator instance
        """
        if not params:
            return self
        valid_params = self._get_params(return_none=True)

        for key, value in params.items():
            if key not in valid_params:
                raise HnswException(
                    'Invalid parameter %s for HnswEstimator. '
                    'Check the list of available parameters '
                    'with `get_params().keys()`.'
                )
                setattr(self, key, value)

        return self

    @property
    def effective_metric_(self):
        """
        Returns
        -------
        Distance that should be used for finding nearest vectors.
        """
        return self.distance

    @property
    def n_samples_fit_(self):
        """
        Returns
        -------
        Number of samples in the fitted data.
        """
        self._check_index()
        return self._index_data.shape[0]

    def kneighbors(self, X=None, n_neighbors=None, return_distance=True, search_neighborhood_size=None,
                   distance_calc_limit=0):
        """Finds the approximate K-neighbors of a point.
        Returns indices of and distances to the neighbors of each point.

        Parameters
        ----------
        X : array-like, shape (n_queries, n_features) or None
            The query point or points.
            If not provided, neighbors of each indexed point are returned.
            In this case, the query point is not considered its own neighbor.
        n_neighbors : int, default=None
            Number of neighbors required for each sample. The default is the
            value passed to the constructor.
        return_distance : bool, default=True
            Whether or not to return the distances.

        search_neighborhood_size : int, default=None
            Search neighborhood size for ANN-search.
            Higher values improve search quality in expense of search time.
            It should be equal or greater than top_size.
            If None set to n_neighbors * 2.

        distance_calc_limit : int (default=0)
            Limit of distance calculation.
            To guarantee satisfactory search time at the expense of quality.
            0 is equivalent to no limit.

        Returns
        -------
        neigh_dist :numpy.ndarray of shape (n_queries, n_neighbors)
            Array representing the lengths to points, only present if
            return_distance=True
        neigh_ind : numpy.ndarray of shape (n_queries, n_neighbors)
            Indices of the nearest points in the population matrix.
        """
        self._check_index()

        if X is None:
            X = self._index_data
        if n_neighbors is None:
            n_neighbors = self.n_neighbors
        if search_neighborhood_size is None:
            search_neighborhood_size = n_neighbors * 2

        return self._index._kneighbors(X, n_neighbors, return_distance, self.distance, search_neighborhood_size,
                                       distance_calc_limit)


class OnlineHnsw:
    """
    Class for building and working with Online Hierarchical Navigable Small World index.
    """
    def __init__(self, dtype, dimension, distance, max_neighbors=None, search_neighborhood_size=None, num_vertices=None, level_size_decay=None):
        """
        Create object with given options.

        Parameters
        ----------
        dtype : EVectorComponentType
            Type of vectors.
        dimension : int
            Dimension of vectors.
        distance : EDistance
            Distance that should be used for finding nearest vectors.
        max_neighbors : int (default=32)
            Maximum number of neighbors that every item can be connected with.
        search_neighborhood_size : int (default=300)
            Search neighborhood size for ANN-search.
            Higher values improve search quality in expense of building time.
        num_vertices : int (default=0)
            Expected number of vectors in storage.
        level_size_decay : int (default=max_neighbors/2)
            Base of exponent for decaying level sizes.
        """
        self.dtype = dtype
        self.dimension = dimension
        params = {}
        all_params = ["max_neighbors", "search_neighborhood_size", "num_vertices", "level_size_decay"]
        for key, value in iteritems(locals()):
            if key in all_params and value is not None:
                params[key] = value
        self._online_index = _OnlineHnswDenseVectorIndex[dtype](dimension, distance, json.dumps(params))

    def get_nearest_and_add_item(self, query):
        """
        Get approximate nearest neighbors for query from index and add item to index

        Parameters
        ----------
        query : list or numpy.ndarray
            Vector for which nearest neighbors should be found.
            Vector which should be added in index.

        Returns
        -------
        neighbors : list of tuples (id, distance) with length = search_neighborhood_size
        """
        return self._online_index._get_nearest_neighbors_and_add_item(query)

    def get_nearest(self, query, top_size=0):
        """
        Get approximate nearest neighbors for query from index.

        Parameters
        ----------
        query : list or numpy.ndarray
            Vector for which nearest neighbors should be found.
        top_size : int
            Required number of neighbors.

        Returns
        -------
        neighbors : list of tuples (id, distance)
        """
        return self._online_index._get_nearest_neighbors(query, top_size)

    def add_item(self, item):
        """
        Add item in index.

        Parameters
        ----------
        item : list or numpy.ndarray
            Vector which should be added in index.
        """
        self._online_index._add_item(item)

    def get_item(self, id):
        """
        Get item from storage by id.

        Parameters
        ----------
        id : int
            Index of item in storage.

        Returns
        -------
        item : numpy.ndarray
        """
        return self._online_index._get_item(id)

    def get_num_items(self):
        """
        Get the number of items in storage.

        Returns
        -------
        num_items : int
        """
        return self._online_index._get_num_items()



from . import _catboost
from .core import Pool, CatBoostError, ARRAY_TYPES, PATH_TYPES, STRING_TYPES, fspath, _update_params_quantize_part, _process_synonyms
from collections import defaultdict
from contextlib import contextmanager
import sys
import numpy as np
import warnings

_eval_metric_util = _catboost._eval_metric_util
_get_roc_curve = _catboost._get_roc_curve
_get_confusion_matrix = _catboost._get_confusion_matrix
_select_threshold = _catboost._select_threshold
_NumpyAwareEncoder = _catboost._NumpyAwareEncoder
_get_onnx_model = _catboost._get_onnx_model
_calculate_quantization_grid = _catboost._calculate_quantization_grid

compute_wx_test = _catboost.compute_wx_test
TargetStats = _catboost.TargetStats
DataMetaInfo = _catboost.DataMetaInfo
compute_training_options = _catboost.compute_training_options


@contextmanager
def _import_matplotlib():
    try:
        import matplotlib.pyplot as plt
    except ImportError as e:
        warnings.warn("To draw plots you should install matplotlib.")
        raise ImportError(str(e))
    yield plt


def _draw(plt, x, y, x_label, y_label, title):
    plt.figure(figsize=(16, 8))

    plt.plot(x, y, alpha=0.5, lw=2)

    plt.xticks(fontsize=16)
    plt.yticks(fontsize=16)
    plt.grid(True)
    plt.xlabel(x_label, fontsize=16)
    plt.ylabel(y_label, fontsize=16)
    plt.title(title, fontsize=20)
    plt.show()


def create_cd(
    label=None,
    cat_features=None,
    text_features=None,
    embedding_features=None,
    weight=None,
    baseline=None,
    doc_id=None,
    group_id=None,
    subgroup_id=None,
    timestamp=None,
    auxiliary_columns=None,
    feature_names=None,
    output_path='train.cd'
):
    _from_param_to_cd = {
        'label': 'Label',
        'weight': 'Weight',
        'baseline': 'Baseline',
        'doc_id': 'DocId',
        'group_id': 'GroupId',
        'subgroup_id': 'SubgroupId',
        'timestamp': 'Timestamp'
    }
    _column_description = defaultdict(lambda: ['Num', ''])
    for key, value in locals().copy().items():
        if not (key.startswith('_') or value is None):
            if key in ('cat_features', 'text_features', 'embedding_features', 'auxiliary_columns'):
                if isinstance(value, int):
                    value = [value]
                for index in value:
                    if not isinstance(index, int):
                        raise CatBoostError('Unsupported index type. Expected int, got {}'.format(type(index)))
                    if index in _column_description:
                        raise CatBoostError('The index {} occurs more than once'.format(index))
                    if key == 'cat_features':
                        _column_description[index] = ['Categ', '']
                    elif key == 'text_features':
                        _column_description[index] = ['Text', '']
                    elif key == 'embedding_features':
                        _column_description[index] = ['NumVector', '']
                    else:
                        _column_description[index] = ['Auxiliary', '']
            elif key not in ('feature_names', 'output_path'):
                if not isinstance(value, int):
                    raise CatBoostError('Unsupported index type. Expected int, got {}'.format(type(value)))
                if value in _column_description:
                    raise CatBoostError('The index {} occurs more than once'.format(value))
                _column_description[value] = [_from_param_to_cd[key], '']
    if feature_names is not None:
        for feature_column_index, name in feature_names.items():
            if _column_description[feature_column_index][0] not in ('Num', 'Categ', 'Text', 'NumVector'):
                raise CatBoostError('feature_names contains index {} that does not correspond to feature column'.format(feature_column_index))
            _column_description[feature_column_index][1] = name
    with open(fspath(output_path), 'w') as f:
        for index, (title, name) in sorted(_column_description.items()):
            f.write('{}\t{}\t{}\n'.format(index, title, name))


def read_cd(cd_file, column_count=None, data_file=None, canonize_column_types=False):
    """
    Reads CatBoost column description file
    (see https://catboost.ai/docs/concepts/input-data_column-descfile.html#input-data_column-descfile)

    Parameters
    ----------
    cd_file : str or pathlib.Path
        path to column description file

    column_count : integer
        total number of columns

    data_file : str or pathlib.Path
        path to dataset file in CatBoost format
        specify either column_count directly or data_file to detect it

    canonize_column_types : bool
        if set to True types for columns with synonyms are renamed to canonical type.

    Returns
    -------
    dict with keys:
        "column_type_to_indices" :
            dict of column_type -> column_indices list, column_type is 'Label', 'Categ' etc.

        "column_dtypes" : dict of column_name -> numpy.dtype or 'category'

        "cat_feature_indices" : list of integers
            indices of categorical features in array of all features.
            Note: indices in array of features, not indices in array of all columns!

        "text_feature_indices" : list of integers
            indices of text features in array of all features.
            Note: indices in array of features, not indices in array of all columns!

        "embedding_feature_indices" : list of integers
            indices of embedding features in array of all features.
            Note: indices in array of features, not indices in array of all columns!

        "column_names" : list of strings

        "non_feature_column_indices" : list of integers
    """

    column_type_synonyms_map = {
        'Target': 'Label',
        'DocId': 'SampleId',
        'QueryId': 'GroupId'
    }

    if column_count is None:
        if data_file is None:
            raise Exception(
                'Cannot obtain column count: either specify column_count parameter or specify data_file '
                + 'parameter to get it'
            )
        with open(fspath(data_file)) as f:
            column_count = len(f.readline()[:-1].split('\t'))

    column_type_to_indices = {}
    column_dtypes = {}
    cat_feature_indices = []
    text_feature_indices = []
    embedding_feature_indices = []
    column_names = []
    non_feature_column_indices = []


    def add_missed_columns(start_column_idx, end_column_idx, non_feature_column_count):
        for missed_column_idx in range(start_column_idx, end_column_idx):
            column_name = 'feature_%i' % (missed_column_idx - non_feature_column_count)
            column_names.append(column_name)
            column_type_to_indices.setdefault('Num', []).append(missed_column_idx)
            column_dtypes[column_name] = np.float32

    last_column_idx = -1
    with open(fspath(cd_file)) as f:
        for line_idx, line in enumerate(f):
            line = line.strip()

            # some cd files in the wild contain empty lines
            if len(line) == 0:
                continue

            line_columns = line.split('\t')
            if len(line_columns) not in [2, 3]:
                raise Exception('Wrong number of columns in cd file')

            column_idx = int(line_columns[0])
            if column_idx <= last_column_idx:
                raise Exception('Non-increasing column indices in cd file')

            add_missed_columns(last_column_idx + 1, column_idx, len(non_feature_column_indices))

            column_type = line_columns[1]
            if canonize_column_types:
                column_type = column_type_synonyms_map.get(column_type, column_type)

            column_type_to_indices.setdefault(column_type, []).append(column_idx)

            column_name = None
            if len(line_columns) == 3:
                column_name = line_columns[2]

            if column_type in ['Num', 'Categ', 'Text', 'NumVector']:
                feature_idx = column_idx - len(non_feature_column_indices)
                if column_name is None:
                    column_name = 'feature_%i' % feature_idx
                if column_type == 'Categ':
                    cat_feature_indices.append(feature_idx)
                    column_dtypes[column_name] = 'category'
                elif column_type == 'Text':
                    text_feature_indices.append(feature_idx)
                    column_dtypes[column_name] = object
                elif column_type == 'NumVector':
                    embedding_feature_indices.append(feature_idx)
                    column_dtypes[column_name] = object
                else:
                    column_dtypes[column_name] = np.float32
            else:
                non_feature_column_indices.append(column_idx)
                if column_name is None:
                    column_name = column_type

            column_names.append(column_name)

            last_column_idx = column_idx

    add_missed_columns(last_column_idx + 1, column_count, len(non_feature_column_indices))

    return {
        'column_type_to_indices' : column_type_to_indices,
        'column_dtypes' : column_dtypes,
        'cat_feature_indices' : cat_feature_indices,
        'text_feature_indices' : text_feature_indices,
        'embedding_feature_indices' : embedding_feature_indices,
        'column_names' : column_names,
        'non_feature_column_indices' : non_feature_column_indices
    }


def eval_metric(label, approx, metric, weight=None, group_id=None, group_weight=None, subgroup_id=None, pairs=None, thread_count=-1):
    """
    Evaluate metrics with raw approxes and labels.

    Parameters
    ----------
    label : list or numpy.ndarrays or pandas.DataFrame or pandas.Series
        Object labels with shape (n_objects,) or (n_object, n_target_dimension)

    approx : list or numpy.ndarrays or pandas.DataFrame or pandas.Series
        Object approxes with shape (n_objects,) or (n_object, n_approx_dimension).

    metric : string
        Metric name.

    weight : list or numpy.ndarray or pandas.DataFrame or pandas.Series, optional (default=None)
        Object weights.

    group_id : list or numpy.ndarray or pandas.DataFrame or pandas.Series, optional (default=None)
        Object group ids.

    group_weight : list or numpy.ndarray or pandas.DataFrame or pandas.Series, optional (default=None)
        Group weights.

    subgroup_id : list or numpy.ndarray, optional (default=None)
        subgroup id for each instance.
        If not None, giving 1 dimensional array like data.

    pairs : list or numpy.ndarray or pandas.DataFrame or string or pathlib.Path
        The pairs description.
        If list or numpy.ndarrays or pandas.DataFrame, giving 2 dimensional.
        The shape should be Nx2, where N is the pairs' count. The first element of the pair is
        the index of winner object in the training set. The second element of the pair is
        the index of loser object in the training set.
        If string or pathlib.Path, giving the path to the file with pairs description.

    thread_count : int, optional (default=-1)
        Number of threads to work with.
        If -1, then the number of threads is set to the number of CPU cores.

    Returns
    -------
    metric results : list with metric values.
    """
    if len(label) > 0:
        label = np.transpose(label) if isinstance(label[0], ARRAY_TYPES) else [label]
    if len(approx) == 0:
        approx = [[]]
    approx = np.transpose(approx) if isinstance(approx[0], ARRAY_TYPES) else [approx]
    return _eval_metric_util(label, approx, metric, weight, group_id, group_weight, subgroup_id, pairs, thread_count)


def get_gpu_device_count():
    return _catboost._get_gpu_device_count()


def reset_trace_backend(filename):
    _catboost._reset_trace_backend(filename)


def get_confusion_matrix(model, data, thread_count=-1):
    """
    Build confusion matrix.

    Parameters
    ----------
    model : catboost.CatBoost
        The trained model.

    data : catboost.Pool
        A set of samples to build confusion matrix with.

    thread_count : int (default=-1)
        Number of threads to work with.
        If -1, then the number of threads is set to the number of CPU cores.

    Returns
    -------
    confusion matrix : array, shape = [n_classes, n_classes]
    """
    if not isinstance(data, Pool):
        raise CatBoostError('data must be a catboost.Pool')

    return _get_confusion_matrix(model._object, data, thread_count);


def get_roc_curve(model, data, thread_count=-1, plot=False):
    """
    Build points of ROC curve.

    Parameters
    ----------
    model : catboost.CatBoost
        The trained model.

    data : catboost.Pool or list of catboost.Pool
        A set of samples to build ROC curve with.

    thread_count : int (default=-1)
        Number of threads to work with.
        If -1, then the number of threads is set to the number of CPU cores.

    plot : bool, optional (default=False)
        If True, draw curve.

    Returns
    -------
    curve points : tuple of three arrays (fpr, tpr, thresholds)
    """
    if type(data) == Pool:
        data = [data]
    if not isinstance(data, list):
        raise CatBoostError('data must be a catboost.Pool or list of pools.')
    for pool in data:
        if not isinstance(pool, Pool):
            raise CatBoostError('one of data pools is not catboost.Pool')

    roc_curve = _get_roc_curve(model._object, data, thread_count)

    if plot:
        with _import_matplotlib() as plt:
            _draw(plt, roc_curve[0], roc_curve[1], 'False Positive Rate', 'True Positive Rate', 'ROC Curve')

    return roc_curve


def get_fpr_curve(model=None, data=None, curve=None, thread_count=-1, plot=False):
    """
    Build points of FPR curve.

    Parameters
    ----------
    model : catboost.CatBoost
        The trained model.

    data : catboost.Pool or list of catboost.Pool
        A set of samples to build ROC curve with.

    curve : tuple of three arrays (fpr, tpr, thresholds)
        ROC curve points in format of get_roc_curve returned value.
        If set, data parameter must not be set.

    thread_count : int (default=-1)
        Number of threads to work with.
        If -1, then the number of threads is set to the number of CPU cores.

    plot : bool, optional (default=False)
        If True, draw curve.

    Returns
    -------
    curve points : tuple of two arrays (thresholds, fpr)
    """
    if curve is not None:
        if data is not None:
            raise CatBoostError('Only one of the parameters data and curve should be set.')
        if not (isinstance(curve, list) or isinstance(curve, tuple)) or len(curve) != 3:
            raise CatBoostError('curve must be list or tuple of three arrays (fpr, tpr, thresholds).')
        fpr, thresholds = curve[0][:], curve[2][:]
    else:
        if model is None or data is None:
            raise CatBoostError('model and data parameters should be set when curve parameter is None.')
        fpr, _, thresholds = get_roc_curve(model, data, thread_count)

    if plot:
        with _import_matplotlib() as plt:
            _draw(plt, thresholds, fpr, 'Thresholds', 'False Positive Rate', 'FPR Curve')

    return thresholds, fpr


def get_fnr_curve(model=None, data=None, curve=None, thread_count=-1, plot=False):
    """
    Build points of FNR curve.

    Parameters
    ----------
    model : catboost.CatBoost
        The trained model.

    data : catboost.Pool or list of catboost.Pool
        A set of samples to build ROC curve with.

    curve : tuple of three arrays (fpr, tpr, thresholds)
        ROC curve points in format of get_roc_curve returned value.
        If set, data parameter must not be set.

    thread_count : int (default=-1)
        Number of threads to work with.
        If -1, then the number of threads is set to the number of CPU cores.

    plot : bool, optional (default=False)
        If True, draw curve.

    Returns
    -------
    curve points : tuple of two arrays (thresholds, fnr)
    """
    if curve is not None:
        if data is not None:
            raise CatBoostError('Only one of the parameters data and curve should be set.')
        if not (isinstance(curve, list) or isinstance(curve, tuple)) or len(curve) != 3:
            raise CatBoostError('curve must be list or tuple of three arrays (fpr, tpr, thresholds).')
        tpr, thresholds = curve[1], curve[2][:]
    else:
        if model is None or data is None:
            raise CatBoostError('model and data parameters should be set when curve parameter is None.')
        _, tpr, thresholds = get_roc_curve(model, data, thread_count)
    fnr = np.array([1 - x for x in tpr])

    if plot:
        with _import_matplotlib() as plt:
            _draw(plt, thresholds, fnr, 'Thresholds', 'False Negative Rate', 'FNR Curve')

    return thresholds, fnr


def select_threshold(model=None, data=None, curve=None, FPR=None, FNR=None, thread_count=-1):
    """
    Selects a threshold for prediction.

    Parameters
    ----------
    model : catboost.CatBoost
        The trained model.

    data : catboost.Pool or list of catboost.Pool
        Set of samples to build ROC curve with.
        If set, curve parameter must not be set.

    curve : tuple of three arrays (fpr, tpr, thresholds)
        ROC curve points in format of get_roc_curve returned value.
        If set, data parameter must not be set.

    FPR : desired false-positive rate

    FNR : desired false-negative rate (only one of FPR and FNR should be chosen)

    thread_count : int (default=-1)
        Number of threads to work with.
        If -1, then the number of threads is set to the number of CPU cores.

    Returns
    -------
    threshold : double
    """
    if data is not None:
        if curve is not None:
            raise CatBoostError('Only one of the parameters data and curve should be set.')
        if model is None:
            raise CatBoostError('model and data parameters should be set when curve parameter is None.')
        if type(data) == Pool:
            data = [data]
        if not isinstance(data, list):
            raise CatBoostError('data must be a catboost.Pool or list of pools.')
        for pool in data:
            if not isinstance(pool, Pool):
                raise CatBoostError('one of data pools is not catboost.Pool')
        return _select_threshold(model._object, data, None, FPR, FNR, thread_count)
    elif curve is not None:
        if not (isinstance(curve, list) or isinstance(curve, tuple)) or len(curve) != 3:
            raise CatBoostError('curve must be list or tuple of three arrays (fpr, tpr, thresholds).')
        return _select_threshold(None, None, curve, FPR, FNR, thread_count)
    else:
        raise CatBoostError('One of the parameters data and curve should be set.')


def quantize(
    data_path,
    column_description=None,
    pairs=None,
    delimiter='\t',
    has_header=False,
    ignore_csv_quoting=False,
    feature_names=None,
    thread_count=-1,
    ignored_features=None,
    per_float_feature_quantization=None,
    border_count=None,
    max_bin=None,
    feature_border_type=None,
    nan_mode=None,
    input_borders=None,
    task_type=None,
    used_ram_limit=None,
    random_seed=None,
    log_cout=sys.stdout,
    log_cerr=sys.stderr,
    **kwargs
):
    """
    Construct quantized Pool from non-quantized pool stored in file.
    This method does not load whole non-quantized source dataset into memory
    so it can be used for huge datasets that fit in memory only after quantization.

    Parameters
    ----------
    data_path : string or pathlib.Path
        Path (with optional scheme) to non-quantized dataset.

    column_description : string, [default=None]
        ColumnsDescription parameter.
        There are several columns description types: Label, Categ, Num, Auxiliary, DocId, Weight, Baseline, GroupId, Timestamp.
        All columns are Num as default, it's not necessary to specify
        this type of columns. Default Label column index is 0 (zero).
        If None, Label column is 0 (zero) as default, all data columns are Num as default.
        If string or pathlib.Path, giving the path to the file with ColumnsDescription in column_description format.

    pairs : string or pathlib.Path, [default=None]
        Path to the file with pairs description.

    has_header : bool, [default=False]
        If True, read column names from first line.

    ignore_csv_quoting : bool optional (default=False)
        If True ignore quoting '"'.

    feature_names : string or pathlib.Path, [default=None]
        Path with scheme for feature names data to load.

    thread_count : int, [default=-1]
        Thread count for data processing.
        If -1, then the number of threads is set to the number of CPU cores.

    ignored_features : list, [default=None]
        Indices or names of features that should be excluded when training.

    per_float_feature_quantization : list of strings, [default=None]
        List of float binarization descriptions.
        Format : described in documentation on catboost.ai
        Example 1: ['0:1024'] means that feature 0 will have 1024 borders.
        Example 2: ['0:border_count=1024', '1:border_count=1024', ...] means that two first features have 1024 borders.
        Example 3: ['0:nan_mode=Forbidden,border_count=32,border_type=GreedyLogSum',
                    '1:nan_mode=Forbidden,border_count=32,border_type=GreedyLogSum'] - defines more quantization properties for first two features.

    border_count : int, [default = 254 for training on CPU or 128 for training on GPU]
        The number of partitions in numeric features binarization. Used in the preliminary calculation.
        range: [1,65535] on CPU, [1,255] on GPU

    max_bin : float, synonym for border_count.

    feature_border_type : string, [default='GreedyLogSum']
        The binarization mode in numeric features binarization. Used in the preliminary calculation.
        Possible values:
            - 'Median'
            - 'Uniform'
            - 'UniformAndQuantiles'
            - 'GreedyLogSum'
            - 'MaxLogSum'
            - 'MinEntropy'

    nan_mode : string, [default=None]
        Way to process missing values for numeric features.
        Possible values:
            - 'Forbidden' - raises an exception if there is a missing value for a numeric feature in a dataset.
            - 'Min' - each missing value will be processed as the minimum numerical value.
            - 'Max' - each missing value will be processed as the maximum numerical value.
        If None, then nan_mode=Min.

    input_borders : string or pathlib.Path, [default=None]
        input file with borders used in numeric features binarization.

    task_type : string, [default=None]
        The calcer type used to train the model.
        Possible values:
            - 'CPU'
            - 'GPU'

    used_ram_limit=None

    random_seed : int, [default=None]
        The random seed used for data sampling.
        If None, 0 is used.

    Returns
    -------
    pool : Pool
        Constructed and quantized pool.
    """
    if not data_path:
        raise CatBoostError("Data filename is empty.")
    if not isinstance(data_path, PATH_TYPES):
        raise CatBoostError("Data filename should be string or pathlib.Path type.")

    if pairs is not None and not isinstance(pairs, PATH_TYPES):
        raise CatBoostError("pairs should have None or string or pathlib.Path type when the pool is read from the file.")
    if column_description is not None and not isinstance(column_description, PATH_TYPES):
        raise CatBoostError("column_description should have None or string or pathlib.Path type when the pool is read from the file.")
    if feature_names is not None and not isinstance(feature_names, PATH_TYPES):
        raise CatBoostError("feature_names should have None or string or pathlib.Path type when the pool is read from the file.")

    params = {}
    _process_synonyms(params)

    if border_count is None:
        border_count = max_bin

    if 'dev_block_size' in kwargs:
        params['dev_block_size'] = kwargs.pop('dev_block_size')

    dev_max_subset_size_for_build_borders = kwargs.pop('dev_max_subset_size_for_build_borders', None)

    if kwargs:
        raise CatBoostError("got an unexpected keyword arguments: {}".format(kwargs.keys()))

    _update_params_quantize_part(
        params,
        ignored_features,
        per_float_feature_quantization,
        border_count,
        feature_border_type,
        None, # sparse_features_conflict_fraction
        None, # dev_efb_max_buckets
        nan_mode,
        input_borders,
        task_type,
        used_ram_limit,
        random_seed,
        dev_max_subset_size_for_build_borders
    )

    result = Pool(None)
    result._read(data_path, column_description, pairs, feature_names, delimiter, has_header, ignore_csv_quoting, thread_count, params, log_cout=log_cout, log_cerr=log_cerr)

    return result


def convert_to_onnx_object(model, export_parameters=None, **kwargs):
    """
    Convert given CatBoost model to ONNX-ML model.
    Categorical Features are not supported.

    Parameters
    ----------
    model : CatBoost trained model
    export_parameters : dict [default=None]
        Parameters for ONNX-ML export:
            * onnx_graph_name : string
                The name property of onnx Graph
            * onnx_domain : string
                The domain component of onnx Model
            * onnx_model_version : int
                The model_version component of onnx Model
            * onnx_doc_string : string
                The doc_string component of onnx Model
    Returns
    -------
    onnx_object : ModelProto
        The model in ONNX format
    """
    try:
        import onnx
    except ImportError as e:
        warnings.warn("To get working onnx model you should install onnx.")
        raise ImportError(str(e))

    import json
    if not model.is_fitted():
        raise CatBoostError(
            "There is no trained model to use save_model(). Use fit() to train model. Then use this method.")

    for name, value in kwargs.items():
        if name == 'target_opset' and value not in [None, 2]:
            warnings.warn('target_opset argument is not supported. Default target_opset is 2 (ai.onnx.ml domain)')
        elif name == 'initial_types' and value is not None:
            warnings.warn('initial_types argument is not supported')

    params_string = ""
    if export_parameters:
        params_string = json.dumps(export_parameters, cls=_NumpyAwareEncoder)

    model_str = _get_onnx_model(model._object, params_string)
    onnx_model = onnx.load_model_from_string(model_str)
    return onnx_model

def calculate_quantization_grid(values, border_count, border_type='Median'):
    assert border_count > 0, 'Border count should be > 0'
    return _calculate_quantization_grid(values, border_count, border_type)
